/*****************************************************/
/*          This java file is a part of the          */
/*                                                   */
/*           -  Plouf's Java IRC Client  -           */
/*                                                   */
/*   Copyright (C)  2002 - 2003 Philippe Detournay   */
/*                                                   */
/*         All contacts : theplouf@yahoo.com         */
/*                                                   */
/*  PJIRC is free software; you can redistribute     */
/*  it and/or modify it under the terms of the GNU   */
/*  General Public License as published by the       */
/*  Free Software Foundation; version 2 or later of  */
/*  the License.                                     */
/*                                                   */
/*  PJIRC is distributed in the hope that it will    */
/*  be useful, but WITHOUT ANY WARRANTY; without     */
/*  even the implied warranty of MERCHANTABILITY or  */
/*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   */
/*  General Public License for more details.         */
/*                                                   */
/*  You should have received a copy of the GNU       */
/*  General Public License along with PJIRC; if      */
/*  not, write to the Free Software Foundation,      */
/*  Inc., 59 Temple Place, Suite 330, Boston,        */
/*  MA  02111-1307  USA                              */
/*                                                   */
/*****************************************************/

package irc;

import irc.gui.*;
import irc.gui.pixx.*;
import irc.security.*;
import irc.ident.*;
import java.util.*;
import java.io.*;
import irc.dcc.*;

import java.awt.*;
import java.awt.event.*;

/**
 * The IRC Application. This is the main class of PJIRC.
 */
public class IRCApplication extends IRCObject implements ServerListener,ServerManager,DCCListener,IdentListener,PixxMDIInterfaceListener,WindowListener,AWTSourceListener,ActionListener
{
  private DefaultSource _defaultSource;

  private AWTDefaultSource _awtDefaultSource;
  private Hashtable _status;
  private Hashtable _channels;
  private Hashtable _queries;
  private Hashtable _dccChats;
  private Hashtable _dccFiles;
  private Hashtable _lists;
  private Hashtable _servers;

  private BasicCTCPFilter _filter;
  private Interpretor _inter;

  private IdentWrapper _ident;

  private StartupConfiguration _start;

  private PixxTaskBar _task;
  private PixxMDIInterface _interface;

  private Frame _frame;
  private Frame _aboutFrame;

  private Object _nickLock=new Object();

  /**
   * Create a new IRCApplication.
   * @param config the IRC configuration.
   * @param startupConfig the startup configuration.
   * @param source a container in wich the application will display. Maybe null. If null,
   * a new Frame will be opened.
   * @param helpPage help page. May be null.
   */
  public IRCApplication(IRCConfiguration config,StartupConfiguration startupConfig,Container source)
  {
    super(config);
    _aboutFrame=null;
    _task=new PixxTaskBar(config);
    _start=startupConfig;
    _interface=new PixxMDIInterface(config,_task,true);

    _interface.addPixxMDIInterfaceListener(this);
    if(source==null)
    {
      _frame=new Frame();
      _frame.addWindowListener(this);
      _frame.add(_interface);
      _frame.setFont(new Font("",Font.PLAIN,12));
      _frame.setSize(640,400);
      _frame.show();
    }
    else
    {
      _frame=null;
      source.removeAll();
      source.setLayout(new GridLayout(1,1));
      source.add(_interface);
    }
  }

  /**
   * Init the application.
   */
  public synchronized void init()
  {
    _channels=new Hashtable();
    _queries=new Hashtable();
    _dccChats=new Hashtable();
    _dccFiles=new Hashtable();
    _lists=new Hashtable();
    _status=new Hashtable();
    _servers=new Hashtable();
    _defaultSource=new DefaultSource(_ircConfiguration);
    _filter=new BasicCTCPFilter(_ircConfiguration,this);
    _filter.addDCCListener(this);

    DefaultInterpretor defaultInter=new DefaultInterpretor(_ircConfiguration,_start,this);

    _defaultSource.setInterpretor(defaultInter);
    _inter=new CTCPInterpretor(_ircConfiguration,defaultInter,_filter);
    _inter.addLast(_interface.getInterpretor());

    if(_ircConfiguration.getB("MultiServer")) addAWTDefault();

    if(_ircConfiguration.getB("UseIdentServer"))
    {
  		try
  		{
        _ident=new IdentWrapper(_ircConfiguration);
        Exception e=_ident.start(_start.getName(),this);
  		  if(e!=null)
  		  {
          _defaultSource.report("\3"+"6"+"*** "+getText(TextProvider.IDENT_FAILED_LAUNCH,e.getMessage()));
  		  }
  		}
  		catch(Throwable ex)
  		{
  		}
  	}
  	else
  	{
  	  _ident=null;
    }

    IRCServer server=new IRCServer(_ircConfiguration,_start.getNick(),_start.getAltNick(),_start.getName(),_start.getAlias());
    server.setServers(_start.getHost(),_start.getPort(),_start.getPass());
    newServer(server,_ircConfiguration.getB("AutoConnection"));
  }

  private void updateConnect()
  {
    AWTSource active=_task.getActive();
    if(active==null) return;
    Server currentServer=active.getSource().getServer();
    _interface.setConnected(currentServer.isConnected());
  }

  private void addAWTDefault()
  {
    if(_awtDefaultSource==null)
    {
      _awtDefaultSource=new AWTDefaultSource(_ircConfiguration,_defaultSource);
      _task.addDefaultSource(_awtDefaultSource,true);
    }
  }

  private void removeAWTDefault()
  {
    if(_awtDefaultSource==null) return;
    _task.removeDefaultSource(_awtDefaultSource);
    _awtDefaultSource.release();
    _awtDefaultSource=null;
  }

  public void newServer(Server server,boolean connect)
  {
    server.enumerateSourcesAsCreated(this);
    server.addServerListener(this);
    if(connect) server.connect();
    _servers.put(server,server);
  }

  /**
   * Uninit the application.
   */
  public synchronized void uninit()
  {
    Enumeration en=_servers.keys();
    while(en.hasMoreElements())
    {
      Server s=(Server)en.nextElement();
      s.leave();
    }

    if(_ident!=null) _ident.stop();
		removeAWTDefault();
    _interface.removePixxMDIInterfaceListener(this);
    //_interface.release();
    //_task.release();

    //_defaultSource.release();
    _defaultSource=null;
    //_defaultInter.release();

    /*_status=null;
    _channels=null;
    _queries=null;
    _dccChats=null;
    _dccFiles=null;
    _lists=null;*/

    if(_frame!=null) _frame.removeWindowListener(this);
    _frame=null;
    _filter.removeDCCListener(this);

    if(_aboutFrame!=null)
    {
      _aboutFrame.removeWindowListener(this);
      _aboutFrame.hide();
      _aboutFrame.dispose();
      _aboutFrame=null;
    }
    EventDispatcher.clearCache();
  }

  private void channelCreated(Channel chan,Server s,Boolean bring)
  {
    AWTChannel awt=new AWTChannel(_ircConfiguration,chan);
    awt.addAWTSourceListener(this);
    _task.addChannel(awt,bring.booleanValue());
    _channels.put(chan,awt);
  }

  private void channelRemoved(Channel chan,Server serv)
  {
    AWTChannel s=(AWTChannel)_channels.get(chan);
    s.removeAWTSourceListener(this);
    _task.removeChannel(s);
    _channels.remove(chan);
    s.release();
  }

  private void queryCreated(Query query,Server s,Boolean bring)
  {
    AWTQuery awt=new AWTQuery(_ircConfiguration,query);
    awt.addAWTSourceListener(this);
    _task.addQuery(awt,bring.booleanValue());
    _queries.put(query,awt);
    if(!bring.booleanValue()) _ircConfiguration.getAudioConfiguration().onQuery();
  }

  private void queryRemoved(Query query,Server s)
  {
    AWTQuery q=(AWTQuery)_queries.get(query);
    q.removeAWTSourceListener(this);
    _task.removeQuery(q);
    _queries.remove(query);
    q.release();
  }

  private void statusCreated(Status status,Server s,Boolean bring)
  {
    AWTStatus awt=new AWTStatus(_ircConfiguration,status);
    if(_ircConfiguration.getB("ShowStatus")) _task.addStatus(awt,bring.booleanValue());
    awt.addAWTSourceListener(this);
    _status.put(status,awt);
    if((_status.size()==1) && !_ircConfiguration.getB("multiserver")) removeAWTDefault();
  }

  private void statusRemoved(Status status,Server serv)
  {
    AWTStatus s=(AWTStatus)_status.get(status);
    s.removeAWTSourceListener(this);
    if(_ircConfiguration.getB("ShowStatus")) _task.removeStatus(s);
    _status.remove(status);
    if((_status.size()==0) && !_ircConfiguration.getB("multiserver")) addAWTDefault();
    s.release();
  }

  private void chanListCreated(ChanList list,Server s,Boolean bring)
  {
    AWTChanList cl=new AWTChanList(_ircConfiguration,list);
    _task.addChanList(cl,bring.booleanValue());
    cl.addAWTSourceListener(this);
    _lists.put(list,cl);
  }

  private void chanListRemoved(ChanList list,Server s)
  {
    AWTChanList l=(AWTChanList)_lists.get(list);
    l.removeAWTSourceListener(this);
    _task.removeChanList(l);
    _lists.remove(list);
    l.release();
  }

  private void DCCChatCreated(DCCChat chat,Server s,Boolean bbring)
  {
    boolean bring=bbring.booleanValue();
    AWTDCCChat awt=new AWTDCCChat(_ircConfiguration,chat);
    awt.addAWTSourceListener(this);
    _task.addDCCChat(awt,bring);
    _dccChats.put(chat,awt);
  }

  private void DCCChatRemoved(DCCChat chat,Server s)
  {
    AWTDCCChat c=(AWTDCCChat)_dccChats.get(chat);
    c.removeAWTSourceListener(this);
    _task.removeDCCChat(c);
    _dccChats.remove(chat);
    c.release();
  }

  public void DCCFileCreated(DCCFile file,Server s,Boolean bring)
  {
    _dccFiles.put(file,new AWTDCCFile(_ircConfiguration,file));
  }

  public void DCCFileRemoved(DCCFile file,Server s)
  {
    AWTDCCFile f=(AWTDCCFile)_dccFiles.get(file);
    _dccFiles.remove(file);
    f.close();
    f.release();
  }

  public void sourceCreated(Source source,Server server,Boolean bring)
  {
    source.getInterpretor().addLast(_inter);
    //System.out.println("interpretor added");
    source.setCTCPFilter(_filter);
    //System.out.println("filter set");

    //source.getInterpretor().addLast(_interface.getInterpretor());

    if(source instanceof Channel) channelCreated((Channel)source,server,bring);
    if(source instanceof Query) queryCreated((Query)source,server,bring);
    if(source instanceof Status) statusCreated((Status)source,server,bring);
    if(source instanceof ChanList) chanListCreated((ChanList)source,server,bring);
    if(source instanceof DCCChat) DCCChatCreated((DCCChat)source,server,bring);
    if(source instanceof DCCFile) DCCFileCreated((DCCFile)source,server,bring);
  }

  public void sourceRemoved(Source source,Server server)
  {
    if(source instanceof Channel) channelRemoved((Channel)source,server);
    if(source instanceof Query) queryRemoved((Query)source,server);
    if(source instanceof Status) statusRemoved((Status)source,server);
    if(source instanceof ChanList) chanListRemoved((ChanList)source,server);
    if(source instanceof DCCChat) DCCChatRemoved((DCCChat)source,server);
    if(source instanceof DCCFile) DCCFileRemoved((DCCFile)source,server);
  }

  public void serverLeft(Server s)
  {
    _servers.remove(s);
    s.removeServerListener(this);
  }

  public File DCCFileRequest(String nick,String fileName,Integer size)
  {
    return _ircConfiguration.getSecurityProvider().getSaveFile(getText(TextProvider.FILE_SAVEAS));
  }

  public boolean DCCChatRequest(String nick)
  {
    return true;
  }

  public void serverConnected(Server server)
  {
    updateConnect();
    for(int i=0;i<_start.getCommands().length;i++)
    {
      String cmd=_start.getCommands()[i];
      if((cmd.startsWith("/")) && (server instanceof IRCServer))
        ((IRCServer)server).getStatus().sendString(_start.getCommands()[i]);
      else
        server.execute(_start.getCommands()[i]);
    }
  }

  public void serverDisconnected(Server s)
  {
    updateConnect();
  }

  public void identRequested(String source,Integer result,String reply)
  {
    _defaultSource.report("\3"+"6"+"*** "+getText(TextProvider.IDENT_REQUEST,source));
    String s="";
    switch(result.intValue())
    {
      case IdentListener.IDENT_ERROR:s=getText(TextProvider.IDENT_ERROR);break;
      case IdentListener.IDENT_OK:s=getText(TextProvider.IDENT_REPLIED,reply);break;
      case IdentListener.IDENT_DEFAULT:s=getText(TextProvider.IDENT_REPLIED,getText(TextProvider.IDENT_DEFAULT_USER)+" : "+reply);break;
      case IdentListener.IDENT_NOT_FOUND:s=getText(TextProvider.IDENT_NO_USER);break;
      default: s=getText(TextProvider.IDENT_UNDEFINED);break;
    }
    _defaultSource.report("\3"+"6"+"*** "+s);

  }

  public void identRunning(Integer port)
  {
    _defaultSource.report("\3"+"6"+"*** "+getText(TextProvider.IDENT_RUNNING_ON_PORT,port+""));
  }

  public void identLeaving(String message)
  {
    _defaultSource.report("\3"+"6"+"*** "+getText(TextProvider.IDENT_LEAVING,message));
  }

  public void activeChanged(AWTSource source,PixxMDIInterface mdi)
  {
    if(source!=null)
    {
      if(source.getSource().mayDefault())
        source.getSource().getServer().setDefaultSource(source.getSource());

      _interface.setTitle(source.getTitle(),source.getColorContext());
      if(_frame!=null) _frame.setTitle(source.getStrippedTitle());
    }
    else
    {
      _interface.setTitle("",_ircConfiguration.getDefaultColorContext());
      if(_frame!=null) _frame.setTitle("");
    }
    updateConnect();
  }

  public void titleChanged(AWTSource source)
  {
    if(source!=_interface.getActive()) return;
    _interface.setTitle(source.getTitle(),source.getColorContext());
    if(_frame!=null) _frame.setTitle(source.getStrippedTitle());
  }

  public void eventOccured(AWTSource source)
  {
  }

  public void activated(AWTSource source)
  {
  }

  public void connectTriggered(PixxMDIInterface mdi)
  {
    Server currentServer=_task.getActive().getSource().getServer();
    if(currentServer==null) return;

    if(currentServer.isConnected())
      currentServer.disconnect();
    else
      currentServer.connect();
  }

  private Label createLabel(String text)
  {
    Label b=new Label(text,Label.CENTER);
    b.setFont(new Font("",Font.PLAIN,12));
    return b;
  }

  public void aboutTriggered(PixxMDIInterface mdi)
  {
    if(_aboutFrame!=null) return;
    _aboutFrame=new Frame();
    _aboutFrame.setTitle(getText(TextProvider.ABOUT_ABOUT));
    _aboutFrame.setLayout(new BorderLayout());
    _aboutFrame.setFont(new Font("",Font.PLAIN,12));

    Panel text=new Panel();

    text.setLayout(new GridLayout(19,1));
    text.add(createLabel("Java IRC Client V"+_ircConfiguration.getVersion()));
    text.add(new Panel());
    text.add(createLabel(getText(TextProvider.ABOUT_GPL)));
    text.add(new Panel());
    text.add(createLabel(getText(TextProvider.ABOUT_PROGRAMMING)+" : Philippe Detournay alias Plouf (theplouf@yahoo.com)"));
    text.add(createLabel(getText(TextProvider.ABOUT_DESIGN)+" : Raphael Seegmuller chez pixxservices.com (pixxservices@pixxservices.com)"));
    text.add(new Panel());
    text.add(createLabel(getText(TextProvider.ABOUT_THANKS)));
    text.add(new Panel());
    text.add(createLabel("Mandragor : www.mandragor.org"));
    text.add(createLabel("DWChat : www.dwchat.net"));
    text.add(createLabel("Kombat Falcon.be Jerarckill"));
    text.add(createLabel("Ezequiel"));
    text.add(createLabel("Jiquera"));
    text.add(createLabel("Red Spider"));
    text.add(new Panel());
    text.add(createLabel(getText(TextProvider.ABOUT_SUPPORT)));
    text.add(new Panel());
    text.add(createLabel("http://www.pjirc.com"));
    _aboutFrame.addWindowListener(this);
    _aboutFrame.add(text,"Center");

    _aboutFrame.setSize(500,300);
    _aboutFrame.setResizable(false);
    _aboutFrame.show();
  }

  public void helpTriggered(PixxMDIInterface mdi)
  {
    if(_ircConfiguration.getS("HelpPage")!=null) _ircConfiguration.getURLHandler().openURL(_ircConfiguration.getS("HelpPage"));
  }

  public void windowActivated(WindowEvent e)
  {
  }

  public void windowClosed(WindowEvent e)
  {
    if(e.getSource()==_frame)
    {
      uninit();
      //System.exit(0);
      /*while(true)
      {
        System.gc();
        try
        {
          Thread.sleep(100);
        }
        catch(Exception ex)
        {
        }
      }*/
    }
    else if(e.getSource()==_aboutFrame)
    {
      _aboutFrame.removeWindowListener(this);
      _aboutFrame=null;
    }
  }

  public void windowClosing(WindowEvent e)
  {
    ((Frame)e.getSource()).hide();
    ((Frame)e.getSource()).dispose();
  }

  public void windowDeactivated(WindowEvent e)
  {
  }

  public void windowDeiconified(WindowEvent e)
  {
  }

  public void windowIconified(WindowEvent e)
  {
  }

  public void windowOpened(WindowEvent e)
  {
  }

  private AWTSource getActiveSource()
  {
    return _task.getActive();
  }

  /**
   * Request the active source to gain focus.
   */
  public void requestSourceFocus()
  {
    getActiveSource().requestFocus();
  }

  /**
   * Send the given string to the current source interpretor.
   * @param str string to send to the interpretor.
   */
  public void sendString(String str)
  {
    getActiveSource().getSource().sendString(str);
  }

  /**
   * Set the current textfield text.
   * @param text new textfield text.
   */
  public void setFieldText(String txt)
  {
    getActiveSource().setFieldText(txt);
  }

  /**
   * Get the current textfield text.
   * @return the current textfield text.
   */
  public String getFieldText()
  {
    return getActiveSource().getFieldText();
  }

  /**
   * Validate the current textfield text, as if user pressed return key.
   */
  public void validateText()
  {
    getActiveSource().validateText();
  }

  public String[] cannotUseRequestedNicknames(Server s)
  {
    synchronized(_nickLock)
    {
      Frame f=new Frame();
      f.setLayout(new FlowLayout());
      f.setSize(200,65);
      f.setTitle(getText(TextProvider.GUI_CHANGE_NICK));
      TextField field=new TextField(_start.getNick());
      Button b=new Button("Ok");
      b.addActionListener(this);
      f.add(field);
      f.add(b);
      f.show();
      try
      {
        _nickLock.wait();
      }
      catch(InterruptedException ex)
      {
      }
      f.hide();
      f.remove(b);
      f.remove(field);
      f.dispose();
      String[] ans=new String[1];
      ans[0]=field.getText();
      return ans;
    }
  }

  public void actionPerformed(ActionEvent evt)
  {
    synchronized(_nickLock)
    {
      _nickLock.notifyAll();
    }
  }

  private static void add(Vector v,Object o)
  {
    v.insertElementAt(o,v.size());
  }

  private static void usage()
  {
    System.out.println("Usage :");
    System.out.println("   java irc.IRCApplication -f configfile");
    System.out.println("or java irc.IRCApplication -p nick fullname host port");
    System.out.println("");
    System.out.println("Without any parameter, '-f pjirc.cfg' parameters are assumed.");
  }

  public static void go(String[] args)
  {
    FileHandler file=new LocalFileHandler();
    IRCConfiguration ircConfiguration;
    StartupConfiguration startupConfiguration;

    try
    {

      if((args.length==0) || ((args.length>=2) && (args[0].equals("-f"))))
      {
        String f="pjirc.cfg";
        if(args.length>=2) f=args[1];
        StreamParameterProvider provider=new StreamParameterProvider(file.getInputStream(f));
        ConfigurationLoader loader=new ConfigurationLoader(provider,new NullURLHandler(),new AWTImageLoader(),new NullSoundHandler(),file);
        ircConfiguration=loader.loadIRCConfiguration();
        startupConfiguration=loader.loadStartupConfiguration();
      }
      else if((args.length>=5) && (args[0].equals("-p")))
      {
        StreamParameterProvider provider=new StreamParameterProvider(null);
        ConfigurationLoader loader=new ConfigurationLoader(provider,new NullURLHandler(),new AWTImageLoader(),new NullSoundHandler(),file);
        ircConfiguration=loader.loadIRCConfiguration();
        startupConfiguration=new StartupConfiguration(args[1],"",args[2],new String[] {""},new String[] {args[3]},new int[] {new Integer(args[4]).intValue()},"",new String[] {});;
      }
      else
      {
        usage();
        return;
      }

      IRCApplication application=new IRCApplication(ircConfiguration,startupConfiguration,null);
      application.init();
    }
    catch(Exception ex)
    {
      System.out.println("Error : "+ex.getMessage());
      ex.printStackTrace();
    }
  }

  public static void main(String[] args)
  {
    go(args);

/*      while(true)
      {
        System.gc();
        try
        {
          Thread.sleep(100);
        }
        catch(Exception ex)
        {
        }
      }*/

  }

}
